<?php

namespace FiloBlu\ThresholdTaxes\Model\ResourceModel;

use Magento\Checkout\Model\Session;
use Magento\Framework\DB\Select;
use Magento\Framework\Model\ResourceModel\Db\Context;
use Magento\Framework\Webapi\Rest\Request;
use Magento\Quote\Model\ResourceModel\Quote\CollectionFactory;
use Magento\Store\Model\StoreManagerInterface;
use Magento\Tax\Helper\Data;
use Magento\Quote\Model\QuoteIdMaskFactory;
use FiloBlu\ThresholdTaxes\Helper\Data as ThresholdHelper;

/**
 *
 */
class Calculation extends \Magento\Tax\Model\ResourceModel\Calculation
{
    /**
     * @var QuoteIdMaskFactory
     */
    protected $quoteIdMaskFactory;
    /**
     * @var Collection
     */
    protected $quoteCollectionFactory;
    /**
     * @var ThresholdHelper
     */
    protected $thresholdHelper;

    /**
     * @var \Magento\Checkout\Model\Session
     */
    private $checkoutSession;
    /**
     * @var \Magento\Framework\Webapi\Rest\Request
     */
    private $request;
    /**
     * @var \Magento\Quote\Model\MaskedQuoteIdToQuoteIdInterface
     */
    private $maskedQuoteIdToQuoteId;

    /**
     * @param Context $context
     * @param Data $taxData
     * @param StoreManagerInterface $storeManager
     * @param Session $checkoutSession
     * @param Request $request
     * @param QuoteIdMaskFactory $quoteIdMaskFactory
     * @param CollectionFactory $quoteCollectionFactory
     * @param null $connectionName
     */
    public function __construct(
        Context $context,
        Data $taxData,
        StoreManagerInterface $storeManager,
        Session $checkoutSession,
        Request $request,
        QuoteIdMaskFactory $quoteIdMaskFactory,
        CollectionFactory $quoteCollectionFactory,
        ThresholdHelper $thresholdHelper,
        $connectionName = null
    ) {
        $this->checkoutSession = $checkoutSession;
        $this->request = $request;
        $this->quoteIdMaskFactory = $quoteIdMaskFactory;
        $this->quoteCollectionFactory = $quoteCollectionFactory;
        $this->thresholdHelper = $thresholdHelper;
        parent::__construct($context, $taxData, $storeManager, $connectionName);
    }

    /**
     * Returns tax rates for request - either pereforms SELECT from DB, or returns already cached result
     * Notice that productClassId due to optimization can be array of ids
     *
     * @param \Magento\Framework\DataObject $request
     * @return array
     * @throws \Magento\Framework\Exception\LocalizedException
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     * @throws \Zend_Db_Select_Exception
     * @SuppressWarnings(PHPMD.CyclomaticComplexity)
     * @SuppressWarnings(PHPMD.NPathComplexity)
     * @SuppressWarnings(PHPMD.ExcessiveMethodLength)
     */
    protected function _getRates($request)
    {
        // Extract params that influence our SELECT statement and use them to create cache key
        $storeId = $this->_storeManager->getStore($request->getStore())->getId();
        $customerClassId = $request->getCustomerClassId();
        $countryId = $request->getCountryId();
        $regionId = $request->getRegionId();
        $postcode = $request->getPostcode();

        // Process productClassId as it can be array or usual value. Form best key for cache.
        $productClassId = $request->getProductClassId();
        $ids = is_array($productClassId) ? $productClassId : [$productClassId];
        foreach ($ids as $key => $val) {
            $ids[$key] = (int)$val; // Make it integer for equal cache keys even in case of null/false/0 values
        }
        $ids = array_unique($ids);
        sort($ids);
        $productClassKey = implode(',', $ids);

        // Form cache key and either get data from cache or from DB
        $cacheKey = implode(
            '|',
            [$storeId, $customerClassId, $productClassKey, $countryId, $regionId, $postcode]
        );

        if (!isset($this->_ratesCache[$cacheKey])) {
            // Make SELECT and get data
            $select = $this->getConnection()->select();
            $select->from(
                ['main_table' => $this->getMainTable()],
                [
                    'tax_calculation_rate_id',
                    'tax_calculation_rule_id',
                    'customer_tax_class_id',
                    'product_tax_class_id'
                ]
            )->where(
                'customer_tax_class_id = ?',
                (int)$customerClassId
            );
            if ($productClassId) {
                $select->where('product_tax_class_id IN (?)', $productClassId);
            }
            $ifnullTitleValue = $this->getConnection()->getCheckSql(
                'title_table.value IS NULL',
                'rate.code',
                'title_table.value'
            );
            $ruleTableAliasName = $this->getConnection()->quoteIdentifier('rule.tax_calculation_rule_id');
            $select->join(
                ['rule' => $this->getTable('tax_calculation_rule')],
                $ruleTableAliasName . ' = main_table.tax_calculation_rule_id',
                ['rule.priority', 'rule.position', 'rule.calculate_subtotal']
            )->join(
                ['rate' => $this->getTable('tax_calculation_rate')],
                'rate.tax_calculation_rate_id = main_table.tax_calculation_rate_id',
                [
                    'value' => 'rate.rate',
                    'tax_country_id' => 'rate.tax_country_id',
                    'tax_region_id' => 'rate.tax_region_id',
                    'tax_postcode' => 'rate.tax_postcode',
                    'tax_calculation_rate_id' => 'rate.tax_calculation_rate_id',
                    'code' => 'rate.code',
                    'from_total'=> new \Zend_Db_Expr('COALESCE (rate.from_total,0)'),
                    'to_total' => new \Zend_Db_Expr('COALESCE (rate.to_total,9999999999999)')
                ]
            )->joinLeft(
                ['title_table' => $this->getTable('tax_calculation_rate_title')],
                'rate.tax_calculation_rate_id = title_table.tax_calculation_rate_id ' .
                "AND title_table.store_id = '{$storeId}'",
                ['title' => $ifnullTitleValue]
            )->where(
                'rate.tax_country_id = ?',
                $countryId
            )->where(
                'rate.tax_region_id IN(?)',
                [0, (int)$regionId]
            );
            $postcodeIsNumeric = is_numeric($postcode);
            $postcodeIsRange = false;
            $originalPostcode = null;
            if (is_string($postcode) && preg_match('/^(.+)-(.+)$/', $postcode, $matches)) {
                if ($countryId == self::USA_COUNTRY_CODE && is_numeric($matches[2]) && strlen($matches[2]) == 4) {
                    $postcodeIsNumeric = true;
                    $originalPostcode = $postcode;
                    $postcode = $matches[1];
                } else {
                    $postcodeIsRange = true;
                    $zipFrom = $matches[1];
                    $zipTo = $matches[2];
                }
            }

            if ($postcodeIsNumeric || $postcodeIsRange) {
                $selectClone = clone $select;
                $selectClone->where('rate.zip_is_range IS NOT NULL');
            }
            $select->where('rate.zip_is_range IS NULL');

            if ($postcode !== '*' || $postcodeIsRange) {
                $select->where(
                    "rate.tax_postcode IS NULL OR rate.tax_postcode IN('*', '', ?)",
                    $postcodeIsRange ? $postcode : $this->_createSearchPostCodeTemplates($postcode, $originalPostcode)
                );
                if ($postcodeIsNumeric) {
                    $selectClone->where('? BETWEEN rate.zip_from AND rate.zip_to', $postcode);
                } elseif ($postcodeIsRange) {
                    $selectClone->where('rate.zip_from >= ?', $zipFrom)
                        ->where('rate.zip_to <= ?', $zipTo);
                }
            }

            /**
             * @see ZF-7592 issue http://framework.zend.com/issues/browse/ZF-7592
             */
            if ($postcodeIsNumeric || $postcodeIsRange) {
                $select = $this->getConnection()->select()->union(
                    ['(' . $select . ')', '(' . $selectClone . ')']
                );
            }

            $select->order(
                'priority ' . Select::SQL_ASC
            )->order(
                'tax_calculation_rule_id ' . Select::SQL_ASC
            )->order(
                'tax_country_id ' . Select::SQL_DESC
            )->order(
                'tax_region_id ' . Select::SQL_DESC
            )->order(
                'tax_postcode ' . Select::SQL_DESC
            )->order(
                'value ' . Select::SQL_DESC
            );

            $fetchResult = $this->getConnection()->fetchAll($select);
            $filteredRates = [];
            if ($fetchResult) {
                foreach ($fetchResult as $rate) {
                    if (!isset($filteredRates[$rate['tax_calculation_rate_id']])) {
                        $filteredRates[$rate['tax_calculation_rate_id']] = $rate;
                    }
                }
            }
            $this->_ratesCache[$cacheKey] = array_values($filteredRates);
        }

        // WallaceIT: Filter out rates not in threshold
        //$objectManager = \Magento\Framework\App\ObjectManager::getInstance();
        //$logger = $objectManager->get('Psr\Log\LoggerInterface');

        if (!$this->thresholdHelper->isEnabled()) {
            return $this->_ratesCache[$cacheKey];
        }

        $quote = $this->getQuote();

        if (!$quote) {
            return $this->_ratesCache[$cacheKey];
        }

        //return $this->_ratesCache[$cacheKey];


        if ($quote->hasItems()) {
            $rateToReturn = [];
            $subtotal = $quote->getTotals()['subtotal']->getValue();

            //$logger->info("Subtotal: ".$subtotal);

            $rates = $this->_ratesCache[$cacheKey];

            foreach ($rates as $key => $rate) {
                if ($rate['from_total'] === null || $rate['to_total'] === null) {
                    continue;
                }

                if ($rate['to_total'] <= 0) {
                    continue;
                }

                if ($subtotal < $rate['from_total'] || $subtotal > $rate['to_total']) {
                    unset($rates[$key]);
                    continue;
                }

                $rateToReturn[] = $rate;
            }

            return $rateToReturn;
        }

        return $this->_ratesCache[$cacheKey];
    }

    /**
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function getQuote()
    {
        if ($this->checkoutSession->getQuoteId()) {
            return $this->checkoutSession->getQuote();
        }

        $maskedQuoteId = $this->request->getParam('cartId');

        if (!$maskedQuoteId) {
            return null;
        }

        if (is_numeric($maskedQuoteId)) {
            $quoteId = $maskedQuoteId;
        } else {
            $quoteMask = $this->quoteIdMaskFactory->create()->load($maskedQuoteId, 'masked_id');
            $quoteId = $quoteMask->getQuoteId();
        }

        return $this->quoteCollectionFactory->create()->addFieldToFilter('entity_id', $quoteId)->getFirstItem();
    }

}
